/* scherzo - Music notation training
 *
 * Copyright © 2010 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <asoundlib.h>

#include "score.h"
#include "mnemon.h"

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof(arr[0]))

#define unused(foo) foo __attribute__((unused))

#define MIDI_BUF_SIZE 4096

typedef struct challenge
{
    bin_t *bin;
    int item_index;
    score_staff_t *staff;
    score_note_t *note;

    int satisfied;
    int mistaken;
} challenge_t;

typedef struct scherzo
{
    void *ctx;

    GtkWidget *window;
    score_t *score;
    int staff_height;
    score_staff_t *treble;
    score_staff_t *bass;
    score_chord_t *chord;
    int midi_fd;
    snd_midi_event_t *snd_midi_event;

    mnemon_t mnemon;
    challenge_t challenge;

    int num_notes_pressed;
    score_note_t **notes_pressed;
} scherzo_t;

/* Forward declarations. */
static score_note_t *
scherzo_add_note (scherzo_t *scherzo, score_pitch_t pitch, int octave);

static void
scherzo_remove_note (scherzo_t *scherzo, score_pitch_t pitch, int octave);

static void
_judge_note (scherzo_t *scherzo, score_note_t *note);

static void
_score_challenge (scherzo_t *scherzo);

static int
on_delete_event_quit (unused (GtkWidget *widget),
		      unused (GdkEvent *event),
		      unused (gpointer user_data))
{
    gtk_main_quit ();

    /* Returning FALSE allows the default handler for delete-event
     * to proceed to cleanup the widget. */
    return FALSE;
}

static int
on_expose_event_draw (GtkWidget *widget,
		      unused (GdkEventExpose *expose),
		      void * user_data)
{
    scherzo_t *scherzo = user_data;
    score_t *score = scherzo->score;
    cairo_t *cr;
    GtkAllocation allocation;
    static const int pad = 10;
    int widget_width;

    gtk_widget_get_allocation (widget, &allocation);
    widget_width = allocation.width;

    cr = gdk_cairo_create (widget->window);

    /* White background */
    cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
    cairo_paint (cr);

    /* Add some padding on the sides and top */
    cairo_translate (cr, pad, scherzo->staff_height);
    score_set_staff_height (score, scherzo->staff_height);
    score_set_width (score, widget_width - 2 * pad);

    score_draw (score, cr);
 
    return TRUE;
}

static int
on_key_press_event (GtkWidget *widget,
		    GdkEventKey *key,
		    void *user_data)
{
    scherzo_t *scherzo = user_data;
    int octave;
    score_pitch_t pitch;

    if (scherzo->challenge.note)
	octave = scherzo->challenge.note->octave;
    else
	octave = 4;

    switch (key->keyval) {
    case GDK_KEY_plus:
    case GDK_KEY_KP_Add:
    case GDK_KEY_equal:
    case GDK_KEY_KP_Equal:
	scherzo->staff_height += 4;
	gtk_widget_queue_draw (widget);
	return TRUE;
	break;
    case GDK_KEY_minus:
    case GDK_KEY_KP_Subtract:
	scherzo->staff_height -= 4;
	gtk_widget_queue_draw (widget);
	return TRUE;
	break;
    case GDK_KEY_q:
    case GDK_KEY_Q:
    case GDK_KEY_Escape:
	gtk_main_quit ();
	return FALSE;
    case GDK_KEY_c:
    case GDK_KEY_C:
	pitch = SCORE_PITCH_C;
	break;
    case GDK_KEY_d:
    case GDK_KEY_D:
	pitch = SCORE_PITCH_D;
	break;
    case GDK_KEY_e:
    case GDK_KEY_E:
	pitch = SCORE_PITCH_E;
	break;
    case GDK_KEY_f:
    case GDK_KEY_F:
	pitch = SCORE_PITCH_F;
	break;
    case GDK_KEY_g:
    case GDK_KEY_G:
	pitch = SCORE_PITCH_G;
	break;
    case GDK_KEY_a:
    case GDK_KEY_A:
	pitch = SCORE_PITCH_A;
	break;
    case GDK_KEY_b:
    case GDK_KEY_B:
	pitch = SCORE_PITCH_B;
	break;
    }

    if ((key->keyval >= GDK_KEY_A && key->keyval <= GDK_KEY_G) ||
	(key->keyval >= GDK_KEY_a && key->keyval <= GDK_KEY_g))
    {
	score_note_t *note;

	note = scherzo_add_note (scherzo, pitch, octave);
	_judge_note (scherzo, note);
	gtk_widget_queue_draw (scherzo->window);

	return TRUE;
    }


    /* Allow an unhandled event to propagate to other handlers. */
    return FALSE;
}

static int
on_key_release_event (unused (GtkWidget *widget),
		      GdkEventKey *key,
		      void *user_data)
{
    scherzo_t *scherzo = user_data;
    int octave;
    score_pitch_t pitch;

    if (scherzo->challenge.note)
	octave = scherzo->challenge.note->octave;
    else
	octave = 4;

    switch (key->keyval) {
    case GDK_KEY_c:
    case GDK_KEY_C:
	pitch = SCORE_PITCH_C;
	break;
    case GDK_KEY_d:
    case GDK_KEY_D:
	pitch = SCORE_PITCH_D;
	break;
    case GDK_KEY_e:
    case GDK_KEY_E:
	pitch = SCORE_PITCH_E;
	break;
    case GDK_KEY_f:
    case GDK_KEY_F:
	pitch = SCORE_PITCH_F;
	break;
    case GDK_KEY_g:
    case GDK_KEY_G:
	pitch = SCORE_PITCH_G;
	break;
    case GDK_KEY_a:
    case GDK_KEY_A:
	pitch = SCORE_PITCH_A;
	break;
    case GDK_KEY_b:
    case GDK_KEY_B:
	pitch = SCORE_PITCH_B;
	break;
    }

    if ((key->keyval >= GDK_KEY_A && key->keyval <= GDK_KEY_G) ||
	(key->keyval >= GDK_KEY_a && key->keyval <= GDK_KEY_g))
    {
	scherzo_remove_note (scherzo, pitch, octave);
	_score_challenge (scherzo);
	gtk_widget_queue_draw (scherzo->window);

	return TRUE;
    }


    /* Allow an unhandled event to propagate to other handlers. */
    return FALSE;
}

static unsigned char
_score_pitch_and_octave_to_midi (score_pitch_t pitch,
				 int octave)
{
    unsigned char midi_note = 12 * (octave + 1);

    switch (SCORE_PITCH_NAME (pitch)) {
    case SCORE_PITCH_NAME_C:
	break;
    case SCORE_PITCH_NAME_D:
	midi_note += 2;
	break;
    case SCORE_PITCH_NAME_E:
	midi_note += 4;
	break;
    case SCORE_PITCH_NAME_F:
	midi_note += 5;
	break;
    case SCORE_PITCH_NAME_G:
	midi_note += 7;
	break;
    case SCORE_PITCH_NAME_A:
	midi_note += 9;
	break;
    case SCORE_PITCH_NAME_B:
	midi_note += 11;
	break;
    }

    switch (SCORE_PITCH_ACCIDENTAL (pitch)) {
    case SCORE_PITCH_ACCIDENTAL_DOUBLE_FLAT:
	midi_note -= 2;
	break;
    case SCORE_PITCH_ACCIDENTAL_FLAT:
	midi_note -= 1;
	break;
    case SCORE_PITCH_ACCIDENTAL_NATURAL:
	break;
    case SCORE_PITCH_ACCIDENTAL_SHARP:
	midi_note += 1;
	break;
    case SCORE_PITCH_ACCIDENTAL_DOUBLE_SHARP:
	midi_note += 2;
	break;
    }

    return midi_note;
}

static void
_midi_to_score_pitch_and_octave (unsigned char midi_note,
				 score_pitch_t *pitch,
				 int *octave)
{
    *octave = midi_note / 12 - 1;

    switch (midi_note % 12)
    {
    case 0:
	*pitch = SCORE_PITCH_C;
	break;
    case 1:
	*pitch = SCORE_PITCH_Cs;
	break;
    case 2:
	*pitch = SCORE_PITCH_D;
	break;
    case 3:
	*pitch = SCORE_PITCH_Ds;
	break;
    case 4:
	*pitch = SCORE_PITCH_E;
	break;
    case 5:
	*pitch = SCORE_PITCH_F;
	break;
    case 6:
	*pitch = SCORE_PITCH_Fs;
	break;
    case 7:
	*pitch = SCORE_PITCH_G;
	break;
    case 8:
	*pitch = SCORE_PITCH_Gs;
	break;
    case 9:
	*pitch = SCORE_PITCH_A;
	break;
    case 10:
	*pitch = SCORE_PITCH_As;
	break;
    case 11:
	*pitch = SCORE_PITCH_B;
	break;
    }
}

/* Determine a chord name (if any) from the current notes pressed */

typedef struct analyzed_note {
    /* Original note being analzyed. */
    score_note_t *note;

    /* Absolute pitch (expressed as midi number). */
    int midi_pitch;

    /* Pitch relative to bass note. */
    int relative_pitch;
} analyzed_note_t;

static int
_compare_analyzed_note (const void *va, const void *vb)
{
    const analyzed_note_t *a = va, *b = vb;

    return a->midi_pitch - b->midi_pitch;
}

static int
_chord_signature_matches (analyzed_note_t *notes,
			  int num_notes,
			  int *signature_pitches,
			  int num_signature_pitches)
{
    int n, s;

    for (n = 0, s = 0; s < num_signature_pitches; s++) {
	if (n >= num_notes)
	    return 0;
	if (notes[n].relative_pitch != signature_pitches[s])
	    return 0;
	n++;
	/* Skip repeated notes in chord being tested. */
	while (n < num_notes &&
	       notes[n].relative_pitch == signature_pitches[s])
	{
	    n++;
	}
    }

    return 1;
}

static void
scherzo_analyze_chord (scherzo_t *scherzo)
{
    void *local = talloc_new (NULL);
    analyzed_note_t *notes;
    unsigned i, num_notes = scherzo->num_notes_pressed;
    int bass_pitch;
    const char *chord_name = NULL;

    struct { int pitches[3]; const char *name; } triads[] = {
	{ {0, 4, 8}, "Augmented triad" },
	{ {0, 4, 7}, "Major triad" },
	{ {0, 3, 7}, "Minor triad" },
	{ {0, 3, 6}, "Diminished triad" }
    };

    if (scherzo->chord) {
	score_remove_chord (scherzo->chord);
	scherzo->chord = NULL;
    }

    if (num_notes == 0)
	goto DONE;

    notes = talloc_array (local, analyzed_note_t, num_notes);
    if (notes == NULL)
	goto DONE;

    for (i = 0; i < num_notes; i++) {
	score_note_t *note = scherzo->notes_pressed[i];
	notes[i].note = note;
	notes[i].midi_pitch = _score_pitch_and_octave_to_midi (note->pitch,
							       note->octave);
	/* Relative pitch will be filled in after sorting. */
	notes[i].relative_pitch = 0;
    }

    qsort (notes, num_notes, sizeof (analyzed_note_t), _compare_analyzed_note);
    
    bass_pitch = notes[0].midi_pitch;

    for (i = 0; i < num_notes; i++) {
	notes[i].relative_pitch = notes[i].midi_pitch - bass_pitch;
    }

    for (i = 0; i < ARRAY_SIZE (triads); i++) {
	if (_chord_signature_matches (notes, num_notes, triads[i].pitches, 3))
	    chord_name = triads[i].name;
    }

/*
    for (i = 0; i < num_sevenths; i++) {
	if (_chord_signature_matches (notes, num_notes, seventh[i].pitches, 4))
	    chord_name = seventh[i].name;
    }
*/

    if (chord_name)
	scherzo->chord = score_add_chord (scherzo->treble, chord_name);
    else
	scherzo->chord = score_add_chord (scherzo->treble, "Unknown or not a chord");

DONE:
    talloc_free (local);
}

static score_note_t *
scherzo_add_note (scherzo_t *scherzo, score_pitch_t pitch, int octave)
{
    score_staff_t *staff;
    score_note_t *note;
    int i;

    if (scherzo->challenge.note)
	staff = scherzo->challenge.staff;
    else
	staff = scherzo->treble;

    /* Do nothing if this note is already pressed. */
    for (i = 0; i < scherzo->num_notes_pressed; i++) {
	if (scherzo->notes_pressed[i]->pitch == pitch &&
	    scherzo->notes_pressed[i]->octave == octave)
	{
	    return scherzo->notes_pressed[i];
	}
    }

    note = score_add_note (staff, pitch, octave, SCORE_DURATION_WHOLE);

    scherzo->num_notes_pressed++;
    scherzo->notes_pressed = talloc_realloc (scherzo->ctx,
					     scherzo->notes_pressed,
					     score_note_t*,
					     scherzo->num_notes_pressed);
    if (scherzo->notes_pressed == NULL) {
	fprintf (stderr, "Out of memory.\n");
	exit (1);
    }

    scherzo->notes_pressed[scherzo->num_notes_pressed - 1] = note;

    scherzo_analyze_chord (scherzo);

    return note;
}


static void
scherzo_remove_note (scherzo_t *scherzo, score_pitch_t pitch, int octave)
{
    score_note_t *note;
    int i;

    for (i = 0; i < scherzo->num_notes_pressed; i++) {
	note = scherzo->notes_pressed[i];
	if (note->pitch == pitch && note->octave == octave) {
	    score_remove_note (note);
	    if (i < scherzo->num_notes_pressed - 1) {
		memmove (scherzo->notes_pressed + i,
			 scherzo->notes_pressed + i + 1,
			 (scherzo->num_notes_pressed - 1 - i) * sizeof (score_note_t*));
	    }
	    scherzo->num_notes_pressed--;
	    i--;
	}
    }

    scherzo_analyze_chord (scherzo);
}

static score_note_t *
scherzo_add_note_midi (scherzo_t *scherzo, unsigned char midi_note)
{
    score_pitch_t pitch;
    int octave;

    _midi_to_score_pitch_and_octave (midi_note, &pitch, &octave);

    return scherzo_add_note (scherzo, pitch, octave);
}


static void
scherzo_remove_note_midi (scherzo_t *scherzo, unsigned char midi_note)
{
    score_pitch_t pitch;
    int octave;
 
    _midi_to_score_pitch_and_octave (midi_note, &pitch, &octave);

    scherzo_remove_note (scherzo, pitch, octave);
}

void
_select_challenge (scherzo_t *scherzo)
{
    category_t *category_unused;
    bool_t introduced_unused;
    item_t *item;
    challenge_t *challenge = &scherzo->challenge;
    score_pitch_t pitch;
    int octave;
    char *s;

    if (challenge->note) {
	score_remove_note (challenge->note);
	challenge->note = NULL;
    }

    mnemon_select_item (&scherzo->mnemon,
			&challenge->bin,
			&challenge->item_index,
			&category_unused,
			&introduced_unused);

    item = challenge->bin->items[challenge->item_index];

    s = item->challenge;
    if (strncmp (s, "treble:", 7) == 0) {
	s += 7;
	challenge->staff = scherzo->treble;
    } else if (strncmp (s, "bass:", 5) == 0) {
	s += 5;
	challenge->staff = scherzo->bass;
    } else {
	fprintf (stderr,
		 "Malformed staff name: %s (expected 'treble:' or 'bass:')\n",
		 s);
	exit (1);
    }

    switch (*s) {
    case 'C':
	pitch = SCORE_PITCH_VALUE(C, NATURAL);
	break;
    case 'D':
	pitch = SCORE_PITCH_VALUE(D, NATURAL);
	break;
    case 'E':
	pitch = SCORE_PITCH_VALUE(E, NATURAL);
	break;
    case 'F':
	pitch = SCORE_PITCH_VALUE(F, NATURAL);
	break;
    case 'G':
	pitch = SCORE_PITCH_VALUE(G, NATURAL);
	break;
    case 'A':
	pitch = SCORE_PITCH_VALUE(A, NATURAL);
	break;
    case 'B':
	pitch = SCORE_PITCH_VALUE(B, NATURAL);
	break;
    default:
	fprintf (stderr, "Malformed pitch name: %s (expected 'A' - 'G')\n", s);
	exit (1);
    }
    s++;

    if (*s < '0' || *s > '9') {
	fprintf (stderr, "Malformed octave number: %s (expected '0' - '9')\n", s);
	exit (1);
    }

    octave = *s - '0';

    challenge->note = score_add_note (challenge->staff, pitch, octave,
				      SCORE_DURATION_WHOLE);
    challenge->satisfied = 0;
    challenge->mistaken = 0;
}

/* Determine whether the user hit the correct note. */
static void
_judge_note (scherzo_t *scherzo, score_note_t *note)
{
    challenge_t *challenge = &scherzo->challenge;

    if (! scherzo->challenge.note) {
	score_set_note_color_rgb (note, 0.0, 0.0, 0.0); /* black */
	return;
    }

    if (note->pitch == challenge->note->pitch &&
	note->octave == challenge->note->octave)
    {
	challenge->satisfied = 1;
	score_set_note_color_rgb (note, 18/256., 130/256., 28/256.); /* green */
    }
    else
    {
	challenge->mistaken = 1;
	score_set_note_color_rgb (note, 184/256., 4/256., 22/256.); /* red */
    }
}

/* If the user got the right note (eventually), then score it in
 * mnemon and show the next note. */
static void
_score_challenge (scherzo_t *scherzo)
{
    challenge_t *challenge = &scherzo->challenge;

    if (! challenge->note)
	return;

    if (! challenge->satisfied)
	return;

    mnemon_score_item (&scherzo->mnemon, challenge->bin, challenge->item_index,
		       ! challenge->mistaken);

    _select_challenge (scherzo);
}

static int
on_midi_input (unused (GIOChannel *channel),
	       unused (GIOCondition condition),
	       void *user_data)
{
    unsigned char buf[MIDI_BUF_SIZE], *next;
    scherzo_t *scherzo = user_data;
    ssize_t remaining;
    snd_seq_event_t event;
    score_note_t *note;
    int need_redraw = FALSE;

    remaining = read (scherzo->midi_fd, buf, MIDI_BUF_SIZE);

    next = buf;
    while (remaining) {
	long consumed;

	consumed = snd_midi_event_encode (scherzo->snd_midi_event,
					  next, remaining, &event);

	remaining -= consumed;
	next += consumed;

	switch (event.type) {
	case SND_SEQ_EVENT_NONE:
	    /* Incomplete event. Nothing to do. */
	    break;
	case SND_SEQ_EVENT_NOTEON:
	    note = scherzo_add_note_midi (scherzo, event.data.note.note);
	    _judge_note (scherzo, note);
	    need_redraw = TRUE;
	    break;
	case SND_SEQ_EVENT_NOTEOFF:
	    scherzo_remove_note_midi (scherzo, event.data.note.note);
	    _score_challenge (scherzo);
	    need_redraw = TRUE;
	    break;
	case SND_SEQ_EVENT_CLOCK:
	    /* Ignore for now as my piano sends a constant stream of these. */
	    break;
	case SND_SEQ_EVENT_SENSING:
	    /* Ignore for now as my piano sends a constant stream of these. */
	    break;
	default:
	    fprintf (stderr, "Fixme: Do not yet know how to handle MIDI event %d\n",
		     event.type);
	    break;
	}
    }

    if (need_redraw)
	gtk_widget_queue_draw (scherzo->window);

    /* Return TRUE to continue to get called in the future. */
    return TRUE;
}

int
main (int argc, char *argv[])
{
    GtkWidget *drawing_area;
    scherzo_t scherzo;
    int err;

    srand (time (NULL));

    gtk_init (&argc, &argv);

    scherzo.ctx = talloc_new (NULL);

    scherzo.score = score_create (scherzo.ctx);
    scherzo.staff_height = 100;
    score_set_staff_height (scherzo.score, scherzo.staff_height);

    score_add_brace (scherzo.score, 2);
    scherzo.treble = score_add_staff (scherzo.score, SCORE_CLEF_G);
    scherzo.bass = score_add_staff (scherzo.score, SCORE_CLEF_F);

    scherzo.chord = NULL;

    scherzo.num_notes_pressed = 0;
    scherzo.notes_pressed = NULL;

    mnemon_init (&scherzo.mnemon);
    /* XXX: Should create a default file if one cannot be loaded. */
    mnemon_load_category (&scherzo.mnemon, "scherzo-notes");

    scherzo.challenge.note = NULL;
/*
    _select_challenge (&scherzo);
*/

    err = snd_midi_event_new (MIDI_BUF_SIZE, &scherzo.snd_midi_event);
    if (err) {
	fprintf (stderr, "Out of memory.\n");
	return 1;
    }

#define MIDI_DEVICE "/dev/midi1"
    scherzo.midi_fd = open (MIDI_DEVICE, O_RDONLY);
    if (scherzo.midi_fd < 0) {
	printf ("failed to open " MIDI_DEVICE ". Midi input will not be available.\n");
    } else {
	GIOChannel *channel;

	channel = g_io_channel_unix_new (scherzo.midi_fd);
	g_io_channel_set_encoding (channel, NULL, NULL);
	g_io_add_watch (channel, G_IO_IN, on_midi_input, &scherzo);
    }

    scherzo.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    gtk_window_set_default_size (GTK_WINDOW (scherzo.window), 1000, 600);

    g_signal_connect (scherzo.window, "delete-event",
		      G_CALLBACK (on_delete_event_quit), NULL);

    drawing_area = gtk_drawing_area_new ();

    gtk_container_add (GTK_CONTAINER (scherzo.window), drawing_area);

    g_signal_connect (drawing_area, "expose-event",  
		      G_CALLBACK (on_expose_event_draw),
		      &scherzo);

    g_signal_connect (scherzo.window, "key-press-event",
		      G_CALLBACK (on_key_press_event),
		      &scherzo);

    g_signal_connect (scherzo.window, "key-release-event",
		      G_CALLBACK (on_key_release_event),
		      &scherzo);
    
    gtk_widget_show_all (scherzo.window);
    
    gtk_main ();

    mnemon_save (&scherzo.mnemon);

    snd_midi_event_free (scherzo.snd_midi_event);

    talloc_free (scherzo.ctx);

    return 0;
}
